from __future__ import annotations
from abc import ABC, abstractmethod
from ..receive import IReceiveService
from ..util.ConnectionDetails import ConnectionDetails


class IExpireReceiveListener(ABC):
    @abstractmethod
    def on_expire_receive(self, message: bytes, details: ConnectionDetails) -> None:
        """"""


class IExpireReceiveService(IReceiveService, ABC):
    @abstractmethod
    def add_expire_listener(self, listener: IExpireReceiveListener) -> IExpireReceiveService:
        """"""

    @abstractmethod
    def reset(self) -> None:
        """"""



from logging import getLogger
from socket import AF_INET, IPPROTO_UDP, SO_BROADCAST, SOCK_DGRAM, SOL_SOCKET, timeout
try:
    from socket import SO_REUSEPORT

    socket_can_reuse_port = True
except ImportError:
    SO_REUSEPORT = 0
    socket_can_reuse_port = False
from threading import Thread
from typing import Optional, Tuple
from .DiscoveryCommonArguments import DiscoveryCommonArguments
from .DiscoveryRequestMessage import DiscoveryRequestMessage
from .DiscoveryResponseMessage import DiscoveryResponseMessage
from .interfaces import IDiscoveryClientService
from ..constants import buffer_size
from ..log.util import log_method_call
from ..service import IService, IServiceManager
from ..socket_ import ISocket, ISocketFactory
from ..util.argument import AnyArguments, create_value_argument, IArguments
from ..util.Lazy import Lazy
from ..util.Placeholder import Placeholder


class DiscoveryClientArguments(IArguments):
    def __init__(self) -> None:
        self.common = DiscoveryCommonArguments()
        self.ip = Lazy(lambda store: '255.255.255.255')
        self.address = Lazy(lambda store: (self.ip(store), self.common.port(store)))
        self.timeout_seconds = Lazy(lambda store: 2)

    def get_arguments(self) -> AnyArguments:
        return [
            create_value_argument(
                self.timeout_seconds, '--discovery-timeout-seconds', int,
                'Retry discovery after this amount of seconds.'),
            create_value_argument(self.ip, '--discovery-ip', str, 'Where to send discovery broadcast.'),
        ]


class DiscoveryClientService(IDiscoveryClientService, IService):
    def __init__(
            self,
            arguments: DiscoveryClientArguments,
            service_manager: IServiceManager,
            socket_factory: ISocketFactory,
    ) -> None:
        self.arguments = arguments
        self.logger = getLogger(__name__)
        self.server_address: Placeholder[Tuple[str, int]] = Placeholder()
        self.should_run = True
        self.socket: Placeholder[ISocket] = Placeholder()
        self.socket_factory = socket_factory
        self.thread = Thread(target=log_method_call(self.logger, self.run))
        service_manager.add_service(self)

    def get_server_address(self) -> Placeholder[Tuple[str, int]]:
        return self.server_address

    def get_service_name(self) -> str:
        return __name__

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        self.thread.join(timeout_seconds)
        return self.thread.is_alive()

    def run(self) -> None:
        with self.socket.set(self.socket_factory.socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) as socket:
            socket.setsockopt(SOL_SOCKET, SO_BROADCAST, 1)
            if socket_can_reuse_port:
                socket.setsockopt(SOL_SOCKET, SO_REUSEPORT, 1)
            socket.settimeout(self.arguments.timeout_seconds.get())
            while self.should_run:
                try:
                    socket.sendto(DiscoveryRequestMessage.encode(), self.arguments.address.get())
                    message, address = socket.recvfrom(buffer_size)
                    if address is not None and DiscoveryResponseMessage.is_valid(message):
                        self.stop_service()
                        response = DiscoveryResponseMessage.decode(message)
                        self.server_address.set((address[0], response.port))
                except timeout:
                    self.logger.warning('timeout')
                except OSError as os_error:
                    if self.should_run:
                        raise os_error

    def start_service(self) -> None:
        self.thread.start()

    def stop_service(self) -> None:
        self.should_run = False
        socket = self.socket.get_optional_and_clear()
        if socket is not None:
            socket.shutdown_guaranteed()



from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING
from .util import configure_log
from ..util.argument import AnyArguments, create_choice_argument, create_value_argument, IArguments
from ..util.Lazy import Lazy


class LogArguments(IArguments):
    def __init__(self) -> None:
        self.log_file = Lazy(lambda store: None)
        self.log_level = Lazy(lambda store: WARNING)

    def configure_log(self) -> None:
        configure_log(self.log_file.get(), self.log_level.get())

    def get_arguments(self) -> AnyArguments:
        return [
            create_value_argument(
                self.log_file, '--log-file', str, 'Gets created if does not exists. Gets appended if exists.'),
            create_choice_argument(self.log_level, '--log-level', {
                'debug': DEBUG,
                'info': INFO,
                'warn': WARNING,
                'error': ERROR,
                'critical': CRITICAL,
            }, 'Verbosity of the log output.')
        ]



from logging import getLogger
from threading import Event
from typing import Optional
from .interfaces import IStopServerService
from ..receive import IReceiveListener, IReceiveService
from ..service import IStartedService, IServiceManager
from ..util.ConnectionDetails import ConnectionDetails


class StopServerService(IStartedService, IStopServerService, IReceiveListener):
    def __init__(
            self,
            receive_service: IReceiveService,
            service_manager: IServiceManager
    ) -> None:
        self.event = Event()
        self.logger = getLogger(__name__)
        receive_service.add_receive_listener(self)
        service_manager.add_started_service(self)

    def get_service_name(self) -> str:
        return __name__

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        return not self.event.is_set()

    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        self.event.set()

    def wait(self) -> None:
        try:
            self.event.wait()
        except KeyboardInterrupt:
            self.logger.info('terminating . . .')

    def stop_service(self) -> None:
        self.event.set()



from abc import ABC, abstractmethod
from typing import Tuple
from ..util.Placeholder import Placeholder


class IDiscoveryClientService(ABC):
    @abstractmethod
    def get_server_address(self) -> Placeholder[Tuple[str, int]]:
        """"""



from __future__ import annotations
from logging import getLogger
from .CarMessage import CarMessage
from .NetworkStatisticsMessage import NetworkStatisticsMessage
from ..expire.interfaces import IExpireReceiveListener, IExpireReceiveService
from ..gpio.interfaces import IGpioService
from ..receive import IReceiveListener
from ..reliable import IReliableConnectListener, IReliableDisconnectListener, IReliableService
from ..timeout.interfaces import ITimeoutReceiveListener, ITimeoutReceiveService
from ..util.Atomic import Atomic
from ..util.ConnectionDetails import ConnectionDetails


class CarServerService(
    IExpireReceiveListener,
    IReceiveListener,
    IReliableConnectListener,
    IReliableDisconnectListener,
    ITimeoutReceiveListener,
):
    def __init__(
            self,
            expire_service: IExpireReceiveService,
            gpio_service: IGpioService,
            reliable_service: IReliableService,
            timeout_service: ITimeoutReceiveService,
    ) -> None:
        self.expire_service = expire_service.add_expire_listener(self)
        self.gpio_service = gpio_service
        self.logger = getLogger(__name__)
        self.network_statistics = Atomic(NetworkStatisticsMessage(0, 0, 0))
        reliable_service.add_reliable_connect_listener(self).add_reliable_disconnect_listener(self)
        timeout_service.add_timeout_listener(self).add_receive_listener(self)

    def on_expire_receive(self, message: bytes, details: ConnectionDetails) -> None:
        with self.network_statistics as (network_statistics, _):
            network_statistics.expire_receive_count += 1

    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        with self.network_statistics as (network_statistics, _):
            network_statistics.receive_count += 1
        car_message = CarMessage.decode(message)
        self.gpio_service.update(car_message.speed, car_message.steering)

    def on_reliable_connect(self, details: ConnectionDetails) -> None:
        self.expire_service.reset()
        with self.network_statistics as (network_statistics, _):
            network_statistics.expire_receive_count = 0
            network_statistics.receive_count = 0
            network_statistics.timeout_receive_count = 0

    def on_reliable_disconnect(self, details: ConnectionDetails) -> None:
        with self.network_statistics as (network_statistics, _):
            self.logger.info(
                f'Client statistics: receive={network_statistics.receive_count}, '
                f'expire={network_statistics.expire_receive_count}, timeout={network_statistics.timeout_receive_count}')

    def on_timeout_receive(self) -> None:
        self.gpio_service.reset()
        with self.network_statistics as (network_statistics, _):
            network_statistics.timeout_receive_count += 1



from __future__ import annotations
from abc import ABC, abstractmethod
from argparse import ArgumentParser, MetavarTypeHelpFormatter
from sys import version_info
from typing import Any, Callable, Dict, Generic, Iterable, Mapping, Optional, Sequence, Tuple, Type, TypeVar
if version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict
from .checking import check_type
from .Lazy import Lazy

K = TypeVar('K')
T = TypeVar('T')
ParsedType = TypeVar('ParsedType')


def get_key_of_value(haystack: Mapping[K, T], needle: T) -> K:
    for key, value in haystack.items():
        if needle == value:
            return key
    raise RuntimeError(f'Value "{needle}" not found in mapping.')


class AddArgumentKwargs(TypedDict, total=False):
    dest: str
    help: str
    type: Type[Any]


AddArgumentArgs = Tuple[str]


class IArgument(Generic[ParsedType], ABC):
    @abstractmethod
    def add_argument(self, parser: ArgumentParser) -> None:
        """"""

    @abstractmethod
    def get_name(self) -> str:
        """"""

    @abstractmethod
    def set_value(self, value: ParsedType) -> None:
        """"""


AnyArgument = IArgument[Any]
AnyArguments = Iterable[AnyArgument]


class IArguments(ABC):
    def get_arguments(self) -> AnyArguments:
        return []


class Argument(Generic[T, ParsedType], IArgument[ParsedType], IArguments):
    def __init__(
            self,
            parser_add_argument: Callable[[ArgumentParser, AddArgumentArgs, AddArgumentKwargs], None],
            default: Optional[str],
            help_: str,
            lazy: Lazy[T],
            name: str,
            parse_value: Callable[[ParsedType], T],
            type_: Type[ParsedType],
    ) -> None:
        self.parser_add_argument = parser_add_argument
        self.help = help_ if default is None else f'{help_} (default: {default or lazy.get_default()})'
        self.lazy = lazy
        self.name = name
        self.parse_value = parse_value
        self.type = type_

    def add_argument(self, parser: ArgumentParser) -> None:
        kwargs = AddArgumentKwargs()
        if self.name[0] in parser.prefix_chars:
            kwargs['dest'] = self.name
        kwargs['help'] = self.help
        kwargs['type'] = self.type
        self.parser_add_argument(parser, (self.name,), kwargs)

    def get_arguments(self) -> AnyArguments:
        return [self]

    def get_name(self) -> str:
        return self.name

    def set_value(self, value: ParsedType) -> None:
        self.lazy.set(self.parse_value(check_type(value, self.type)))


def create_choice_argument(
        lazy: Lazy[T],
        name: str,
        choices: Mapping[str, T],
        help_: str = '',
        default: str = ''
) -> Argument[T, str]:
    def add_argument(parser: ArgumentParser, args: AddArgumentArgs, kwargs: AddArgumentKwargs) -> None:
        parser.add_argument(*args, **kwargs, choices=choices)

    return Argument(
        add_argument, default or get_key_of_value(choices, lazy.get_default()),
        help_, lazy, name, lambda value: choices[value], str,
    )


def create_value_argument(
        lazy: Lazy[T],
        name: str,
        type_: Type[T],
        help_: str = '',
        default: Optional[str] = '',
) -> Argument[T, T]:
    def add_argument(parser: ArgumentParser, args: AddArgumentArgs, kwargs: AddArgumentKwargs) -> None:
        parser.add_argument(*args, **kwargs)

    return Argument(add_argument, default, help_, lazy, name, lambda value: value, type_)


class ArgumentRegistry:
    def __init__(self) -> None:
        self.parser = ArgumentParser(formatter_class=MetavarTypeHelpFormatter)
        self.arguments: Dict[str, AnyArgument] = dict()

    def add_arguments(self, arguments_list: Iterable[IArguments]) -> None:
        for arguments in arguments_list:
            for argument in arguments.get_arguments():
                self.add_argument(argument)

    def add_argument(self, argument: AnyArgument) -> None:
        name = argument.get_name()
        if name in self.arguments:
            raise RuntimeError(f'Argument {name} already used.')
        self.arguments[name] = argument
        argument.add_argument(self.parser)

    def parse_arguments(self, argument_vector: Optional[Sequence[str]] = None) -> None:
        parsed = vars(self.parser.parse_args(argument_vector))
        parsed_keys = set(parsed.keys())
        for name, value in parsed.items():
            if name not in self.arguments:
                raise RuntimeError(f'Unknown argument {name}.')
            if value is not None:
                self.arguments[name].set_value(value)
            parsed_keys.remove(name)
        if len(parsed_keys) != 0:
            unknown_arguments_string = ', '.join((f'{key}: {parsed[key]}' for key in parsed_keys))
            raise RuntimeError(f'Unknown arguments: {unknown_arguments_string}')


def process_arguments(arguments_list: Iterable[IArguments], argument_vector: Optional[Sequence[str]] = None) -> None:
    argument_registry = ArgumentRegistry()
    argument_registry.add_arguments(arguments_list)
    argument_registry.parse_arguments(argument_vector)



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Callable, Optional
from ..receive import IReceiveService


class ITimeoutReceiveListener(ABC):
    @abstractmethod
    def on_timeout_receive(self) -> None:
        """"""


class ITimeoutReceiveService(IReceiveService, ABC):
    @abstractmethod
    def add_timeout_listener(self, timeout_listener: ITimeoutReceiveListener) -> ITimeoutReceiveService:
        """"""

    @abstractmethod
    def set_timeout(self, timeout_seconds: float) -> None:
        """"""


class ITimeoutSendService(ABC):
    @abstractmethod
    def get_send_count(self) -> int:
        """"""

    @abstractmethod
    def set_and_send_immediately(self, message_callback: Callable[[], Optional[bytes]]) -> Optional[bytes]:
        """"""



from __future__ import annotations
from abc import ABC, abstractmethod


class ILatencyListener(ABC):
    @abstractmethod
    def on_latency_available(self, latency: float) -> None:
        """"""

    @abstractmethod
    def on_latency_timeout(self) -> None:
        """"""


class ILatencyServerService(ABC):
    @abstractmethod
    def add_latency_listener(self, latency_listener: ILatencyListener) -> ILatencyServerService:
        """"""



from socket import socket
from typing import ContextManager, Tuple
from .ClientArguments import ClientArguments
from ..car.CarClientService import CarClientService
from ..clock import IClock
from ..constants import discover_server_ip
from ..discovery.DiscoveryClientService import DiscoveryClientArguments, DiscoveryClientService
from ..discovery.interfaces import IDiscoveryClientService
from ..expire.ExpireSendService import ExpireSendService
from ..gpio.interfaces import IGpioService
from ..gstreamer.GStreamerClientService import GStreamerClientArguments, GStreamerClientService
from ..gstreamer.interfaces import IGStreamerClientService
from ..latency.LatencyClientService import LatencyClientService
from ..log.LogArguments import LogArguments
from ..message import message_types
from ..priority.PrioritySendService import PrioritySendService
from ..receive import IReceiveService
from ..reliable import IReliableReceiveSendService
from ..routed.interfaces import IRoutedReceiveService, IRoutedSendService
from ..routed.RoutedSendService import RoutedSendService
from ..routed.RoutedReceiveService import RoutedReceiveService
from ..send import ISendService
from ..service import IServiceManager
from ..socket_ import ISocketFactory
from ..tcp.TcpClientService import TcpClientService
from ..timeout.interfaces import ITimeoutSendService
from ..timeout.TimeoutSendService import TimeoutSendService
from ..udp.UdpService import UdpService
from ..unreliable import IUnreliableReceiveSendService
from ..util.Lazy import Lazy
from ..util.MultiServiceManager import MultiServiceManager
from ..util.Placeholder import Placeholder
from ..util.RTC import rtc
from ..util.Socket import SocketFactory


class Client:
    def __init__(self) -> None:
        self.client_arguments = ClientArguments()
        self.discovery_arguments = DiscoveryClientArguments()
        self.gstreamer_arguments = GStreamerClientArguments()
        self.log_arguments = LogArguments()

        self.clock: Lazy[IClock] = Lazy(lambda store: rtc)
        self.socket_factory: Lazy[ISocketFactory] = Lazy(lambda store: SocketFactory(socket))

        self.multi_service_manager = Lazy(lambda store: MultiServiceManager())
        self.service_manager: Lazy[IServiceManager] = Lazy(lambda store: self.multi_service_manager(store))

        self.discovery_service_implementation = Lazy(lambda store: DiscoveryClientService(
            self.discovery_arguments, self.service_manager(store), self.socket_factory(store)))
        self.discovery_service: Lazy[IDiscoveryClientService] = Lazy(
            lambda store: self.discovery_service_implementation(store))

        def create_server_address(store: bool) -> Placeholder[Tuple[str, int]]:
            if self.client_arguments.server_address(store)[0] == discover_server_ip:
                return self.discovery_service(store).get_server_address()
            else:
                return Placeholder(self.client_arguments.server_address(store))

        self.server_address = Lazy(create_server_address)

        self.tcp_service = Lazy(lambda store: TcpClientService(
            self.clock(store), self.client_arguments.reconnect_seconds(store), self.server_address(store),
            self.service_manager(store), self.socket_factory(store)))
        self.reliable_service: Lazy[IReliableReceiveSendService] = Lazy(lambda store: self.tcp_service(store))

        self.udp_service = Lazy(lambda store: UdpService(
            self.clock(store), False, self.server_address(store), self.service_manager(store),
            self.socket_factory(store)))
        self.unreliable_service: Lazy[IUnreliableReceiveSendService] = Lazy(lambda store: self.udp_service(store))

        self.reliable_routed_receive_service: Lazy[IRoutedReceiveService] = Lazy(lambda store: RoutedReceiveService(
            self.reliable_service(store)))
        self.reliable_routed_send_service: Lazy[IRoutedSendService] = Lazy(lambda store: RoutedSendService(
            self.reliable_service(store)))
        self.unreliable_routed_receive_service: Lazy[IRoutedReceiveService] = Lazy(lambda store: RoutedReceiveService(
            self.unreliable_service(store)))
        self.unreliable_routed_send_service: Lazy[IRoutedSendService] = Lazy(lambda store: RoutedSendService(
            self.unreliable_service(store)))

        self.car_send_service: Lazy[ISendService] = Lazy(
            lambda store: self.unreliable_routed_send_service(store).create_send_service(message_types.car))

        self.priority_service_implementation = Lazy(lambda store: PrioritySendService(
            self.client_arguments.priority(store), self.car_send_service(store)))
        self.priority_service: Lazy[ISendService] = Lazy(lambda store: self.priority_service_implementation(store))

        self.expire_service_implementation = Lazy(lambda store: ExpireSendService(self.priority_service(store)))
        self.expire_service: Lazy[ISendService] = Lazy(lambda store: self.expire_service_implementation(store))

        self.timeout_service_implementation = Lazy(lambda store: TimeoutSendService(
            self.clock(store), self.expire_service(store), self.service_manager(store),
            self.client_arguments.send_timeout_seconds(store)))
        self.timeout_service: Lazy[ITimeoutSendService] = Lazy(lambda store: self.timeout_service_implementation(store))

        self.car_service_implementation = Lazy(lambda store: CarClientService(
            self.reliable_service(store), self.timeout_service(store)))
        self.car_service: Lazy[IGpioService] = Lazy(lambda store: self.car_service_implementation(store))

        self.gstreamer_receive_service: Lazy[IReceiveService] = Lazy(
            lambda store: self.reliable_routed_receive_service(store).create_receive_service(message_types.gstreamer))
        self.gstreamer_send_service: Lazy[ISendService] = Lazy(
            lambda store: self.reliable_routed_send_service(store).create_send_service(message_types.gstreamer))
        self.gstreamer_service_implementation = Lazy(lambda store: GStreamerClientService(
            self.gstreamer_arguments, self.gstreamer_receive_service(store), self.gstreamer_send_service(store)))
        self.gstreamer_service: Lazy[IGStreamerClientService] = Lazy(
            lambda store: self.gstreamer_service_implementation(store))

        self.latency_receive_service: Lazy[IReceiveService] = Lazy(
            lambda store: self.unreliable_routed_receive_service(store).create_receive_service(message_types.latency))
        self.latency_send_service: Lazy[ISendService] = Lazy(
            lambda store: self.unreliable_routed_send_service(store).create_send_service(message_types.latency))
        self.latency_service = Lazy(lambda store: LatencyClientService(
            self.latency_receive_service(store), self.latency_send_service(store)))

    def use_services(self) -> ContextManager[None]:
        return self.clock.get().use_services(self.service_manager.get())
